import unittest

from xdrone.shared.command import Command, SingleDroneCommand, ParallelDroneCommands, AbstractDroneCommand, \
    RepeatDroneNameException


class TestCommands(unittest.TestCase):

    def test_takeoff(self):
        self.assertEqual(Command.takeoff(), Command.takeoff())
        self.assertEqual("takeoff", Command.takeoff().opcode)
        self.assertEqual([], Command.takeoff().operands)
        self.assertEqual("Command: { opcode: takeoff, operands: [] }", str(Command.takeoff()))

    def test_land(self):
        self.assertEqual(Command.land(), Command.land())
        self.assertEqual("land", Command.land().opcode)
        self.assertEqual([], Command.land().operands)
        self.assertEqual("Command: { opcode: land, operands: [] }", str(Command.land()))

    def test_up(self):
        self.assertEqual(Command.up(1), Command.up(1))
        self.assertEqual("up", Command.up(1).opcode)
        self.assertEqual([1], Command.up(1).operands)
        self.assertEqual("Command: { opcode: up, operands: [1] }", str(Command.up(1)))

    def test_down(self):
        self.assertEqual(Command.down(1), Command.down(1))
        self.assertEqual("down", Command.down(1).opcode)
        self.assertEqual([1], Command.down(1).operands)
        self.assertEqual("Command: { opcode: down, operands: [1] }", str(Command.down(1)))

    def test_left(self):
        self.assertEqual(Command.left(1), Command.left(1))
        self.assertEqual("left", Command.left(1).opcode)
        self.assertEqual([1], Command.left(1).operands)
        self.assertEqual("Command: { opcode: left, operands: [1] }", str(Command.left(1)))

    def test_right(self):
        self.assertEqual(Command.right(1), Command.right(1))
        self.assertEqual("right", Command.right(1).opcode)
        self.assertEqual([1], Command.right(1).operands)
        self.assertEqual("Command: { opcode: right, operands: [1] }", str(Command.right(1)))

    def test_forward(self):
        self.assertEqual(Command.forward(1), Command.forward(1))
        self.assertEqual("forward", Command.forward(1).opcode)
        self.assertEqual([1], Command.forward(1).operands)
        self.assertEqual("Command: { opcode: forward, operands: [1] }", str(Command.forward(1)))

    def test_backward(self):
        self.assertEqual(Command.backward(1), Command.backward(1))
        self.assertEqual("backward", Command.backward(1).opcode)
        self.assertEqual([1], Command.backward(1).operands)
        self.assertEqual("Command: { opcode: backward, operands: [1] }", str(Command.backward(1)))

    def test_rotate_left(self):
        self.assertEqual(Command.rotate_left(1), Command.rotate_left(1))
        self.assertEqual("rotate_left", Command.rotate_left(1).opcode)
        self.assertEqual([1], Command.rotate_left(1).operands)
        self.assertEqual("Command: { opcode: rotate_left, operands: [1] }", str(Command.rotate_left(1)))

    def test_rotate_right(self):
        self.assertEqual(Command.rotate_right(1), Command.rotate_right(1))
        self.assertEqual("rotate_right", Command.rotate_right(1).opcode)
        self.assertEqual([1], Command.rotate_right(1).operands)
        self.assertEqual("Command: { opcode: rotate_right, operands: [1] }", str(Command.rotate_right(1)))

    def test_wait(self):
        self.assertEqual(Command.wait(1), Command.wait(1))
        self.assertEqual("wait", Command.wait(1).opcode)
        self.assertEqual([1], Command.wait(1).operands)
        self.assertEqual("Command: { opcode: wait, operands: [1] }", str(Command.wait(1)))

    def test_eq(self):
        commands1 = [None, Command.takeoff(), Command.land(),
                     Command.up(1), Command.down(1), Command.left(1), Command.right(1), Command.forward(1),
                     Command.backward(1), Command.rotate_left(1), Command.rotate_right(1), Command.wait(1),
                     Command.up(2), Command.down(2), Command.left(2), Command.right(2), Command.forward(2),
                     Command.backward(2), Command.rotate_left(2), Command.rotate_right(2), Command.wait(2)]
        commands2 = [None, Command.takeoff(), Command.land(),
                     Command.up(1), Command.down(1), Command.left(1), Command.right(1), Command.forward(1),
                     Command.backward(1), Command.rotate_left(1), Command.rotate_right(1), Command.wait(1),
                     Command.up(2), Command.down(2), Command.left(2), Command.right(2), Command.forward(2),
                     Command.backward(2), Command.rotate_left(2), Command.rotate_right(2), Command.wait(2)]
        for i in range(len(commands1)):
            for j in range(len(commands2)):
                if i == j:
                    self.assertEqual(commands1[i], commands2[j])
                else:
                    self.assertNotEqual(commands1[i], commands2[j])

    def test_corrupted_command_should_not_affect_correct_type(self):
        command = Command.forward(1)
        corrupted_command = Command.forward(1)
        corrupted_command._opcode = "corrupted"
        self.assertNotEqual(command, corrupted_command)
        self.assertEqual("forward", Command.forward(1).opcode)
        self.assertEqual([1], Command.forward(1).operands)
        self.assertEqual("corrupted", corrupted_command.opcode)
        self.assertEqual([1], corrupted_command.operands)


class TestAbstractDroneCommands(unittest.TestCase):

    def test_get_drones_involved(self):
        with self.assertRaises(NotImplementedError) as context:
            AbstractDroneCommand().get_drones_involved()

    def test_to_command_str(self):
        with self.assertRaises(NotImplementedError) as context:
            AbstractDroneCommand().to_command_str()


class TestSingleDroneCommand(unittest.TestCase):

    def test_properties(self):
        drone_command = SingleDroneCommand("abc", Command.takeoff())
        self.assertEqual("abc", drone_command.drone_name)
        self.assertEqual(Command.takeoff(), drone_command.command)

    def test_get_drones_involved(self):
        drone_command = SingleDroneCommand("abc", Command.takeoff())
        self.assertEqual({"abc"}, drone_command.get_drones_involved())

    def test_to_command_str(self):
        drone_command = SingleDroneCommand("abc", Command.takeoff())
        self.assertEqual("abc.takeoff();", drone_command.to_command_str())
        drone_command = SingleDroneCommand("abc", Command.up(1))
        self.assertEqual("abc.up(1);", drone_command.to_command_str())

    def test_str(self):
        drone_command = SingleDroneCommand("abc", Command.takeoff())
        self.assertEqual("SingleDroneCommand: { drone_name: abc, command: Command: { opcode: takeoff, operands: [] } }",
                         str(drone_command))

    def test_repr(self):
        drone_command = SingleDroneCommand("abc", Command.takeoff())
        self.assertEqual("SingleDroneCommand: { drone_name: abc, command: Command: { opcode: takeoff, operands: [] } }",
                         repr(drone_command))

    def test_eq(self):
        self.assertEqual(SingleDroneCommand("abc", Command.takeoff()), SingleDroneCommand("abc", Command.takeoff()))
        self.assertNotEqual(SingleDroneCommand("", Command.takeoff()), SingleDroneCommand("abc", Command.takeoff()))
        self.assertNotEqual(SingleDroneCommand("abc", Command.takeoff()), SingleDroneCommand("abc", Command.land()))
        self.assertNotEqual(None, SingleDroneCommand("abc", Command.takeoff()))

    def test_immutable(self):
        drone_command = SingleDroneCommand("abc", Command.takeoff())
        drone_name = drone_command.drone_name
        command = drone_command.command
        drone_name += "corrupted"
        command._opcode = "corrupted"
        self.assertEqual("abc", drone_command.drone_name)
        self.assertEqual(Command.takeoff(), drone_command.command)


class TestRepeatDroneNameException(unittest.TestCase):

    def test_properties(self):
        exception = RepeatDroneNameException({"abc"})
        self.assertEqual({"abc"}, exception.repeated_names)

    def test_immutable(self):
        exception = RepeatDroneNameException({"abc"})
        exception.repeated_names.add("corrupted")
        self.assertEqual({"abc"}, exception.repeated_names)


class TestParallelCommands(unittest.TestCase):

    def test_properties(self):
        parallel_commands = ParallelDroneCommands()
        self.assertEqual([], parallel_commands.branches)
        self.assertEqual([], parallel_commands.drones_involved_each_branch)
        parallel_commands = ParallelDroneCommands([[], []])
        self.assertEqual([[], []], parallel_commands.branches)
        self.assertEqual([set(), set()], parallel_commands.drones_involved_each_branch)

    def test_get_drones_involved(self):
        parallel_commands = ParallelDroneCommands()
        self.assertEqual(set(), parallel_commands.get_drones_involved())
        parallel_commands = ParallelDroneCommands([[SingleDroneCommand("DRONE1", Command.takeoff())],
                                                   [SingleDroneCommand("DRONE2", Command.takeoff())],
                                                   [ParallelDroneCommands([
                                                       [SingleDroneCommand("DRONE3", Command.takeoff())],
                                                       [SingleDroneCommand("DRONE4", Command.takeoff())]
                                                   ])]])
        self.assertEqual({"DRONE1", "DRONE2", "DRONE3", "DRONE4"}, parallel_commands.get_drones_involved())

    def test_add(self):
        parallel_commands = ParallelDroneCommands()
        parallel_commands.add([])
        self.assertEqual([[]], parallel_commands.branches)
        self.assertEqual([set()], parallel_commands.drones_involved_each_branch)
        parallel_commands.add([SingleDroneCommand("abc", Command.takeoff())])
        self.assertEqual([[], [SingleDroneCommand("abc", Command.takeoff())]], parallel_commands.branches)
        self.assertEqual([set(), {"abc"}], parallel_commands.drones_involved_each_branch)

    def test_add_already_involved_drones_should_give_error(self):
        parallel_commands = ParallelDroneCommands()
        parallel_commands.add([SingleDroneCommand("DRONE1", Command.takeoff())])
        with self.assertRaises(RepeatDroneNameException) as context:
            parallel_commands.add([SingleDroneCommand("DRONE1", Command.takeoff())])
        self.assertTrue({"DRONE1"}, context.exception.repeated_names)

    def test_init_with_repeated_drones_should_give_error(self):
        with self.assertRaises(RepeatDroneNameException) as context:
            ParallelDroneCommands([
                [SingleDroneCommand("DRONE1", Command.takeoff())],
                [SingleDroneCommand("DRONE1", Command.takeoff())]
            ])
        self.assertTrue({"DRONE1"}, context.exception.repeated_names)

    def test_to_command_str(self):
        parallel_commands = ParallelDroneCommands()
        parallel_commands.add([])
        parallel_commands.add([SingleDroneCommand("abc", Command.takeoff()),
                               SingleDroneCommand("abc", Command.up(1))])
        self.assertEqual("{ } || { abc.takeoff(); abc.up(1); };",
                         parallel_commands.to_command_str())
        outer_parallel_commands = ParallelDroneCommands()
        outer_parallel_commands.add([])
        outer_parallel_commands.add([parallel_commands])
        self.assertEqual("{ } || { { } || { abc.takeoff(); abc.up(1); }; };",
                         outer_parallel_commands.to_command_str())

    def test_str(self):
        parallel_commands = ParallelDroneCommands()
        parallel_commands.add([])
        parallel_commands.add([SingleDroneCommand("abc", Command.takeoff())])
        self.assertEqual("ParallelDroneCommands: { [], " +
                         "[SingleDroneCommand: { drone_name: abc, " +
                         "command: Command: { opcode: takeoff, operands: [] } }] }",
                         str(parallel_commands))

    def test_repr(self):
        parallel_commands = ParallelDroneCommands()
        parallel_commands.add([])
        parallel_commands.add([SingleDroneCommand("abc", Command.takeoff())])
        self.assertEqual("ParallelDroneCommands: { [], " +
                         "[SingleDroneCommand: { drone_name: abc, " +
                         "command: Command: { opcode: takeoff, operands: [] } }] }",
                         repr(parallel_commands))

    def test_eq(self):
        parallel_commands1 = ParallelDroneCommands()
        parallel_commands1.add([])
        parallel_commands1.add([SingleDroneCommand("abc", Command.takeoff())])
        parallel_commands2 = ParallelDroneCommands([[], [SingleDroneCommand("abc", Command.takeoff())]])
        parallel_commands3 = ParallelDroneCommands([[SingleDroneCommand("abc", Command.takeoff())], []])
        self.assertEqual(ParallelDroneCommands(), ParallelDroneCommands())
        self.assertEqual(parallel_commands1, parallel_commands2)
        self.assertNotEqual(parallel_commands1, parallel_commands3)
        self.assertNotEqual(None, parallel_commands1)



import json
import zlib
from base64 import b64encode
from typing import List, Dict

from xdrone.shared.command import AbstractDroneCommand, SingleDroneCommand, ParallelDroneCommands
from xdrone.shared.drone_config import DroneConfig


class SimulationConverter:

    def convert(self, drone_commands: List[AbstractDroneCommand], drone_config_map: Dict[str, DroneConfig]):
        commands = self._convert_commands(drone_commands)
        config = self._convert_config(drone_config_map)
        json_obj = {"config": config, "commands": commands}
        return self._compress_json_obj(json_obj)

    def _convert_config(self, drone_config_map: Dict[str, DroneConfig]):
        configs = []
        for name, config in drone_config_map.items():
            configs.append({"name": name,
                            "init_pos": config.init_position,
                            "speed": config.speed_mps,
                            "rotate_speed": config.rotate_speed_dps,
                            "takeoff_height": config.takeoff_height_meters})
        return configs

    def _convert_commands(self, drone_commands: List[AbstractDroneCommand]):
        sequential = []
        for drone_command in drone_commands:
            if isinstance(drone_command, SingleDroneCommand):
                sequential.append({"type": "single",
                                   "name": drone_command.drone_name,
                                   "action": drone_command.command.opcode,
                                   "values": drone_command.command.operands})
            elif isinstance(drone_command, ParallelDroneCommands):
                sequential.append({"type": "parallel",
                                   "branches": [self._convert_commands(branch) for branch in drone_command.branches]})
        return sequential

    def _compress_json_obj(self, json_obj):
        json_str = json.dumps(json_obj)
        json_bytes = json_str.encode('ascii')
        compressed = zlib.compress(json_bytes)
        compressed_b64 = b64encode(compressed)
        compressed_ascii = compressed_b64.decode('ascii')
        return compressed_ascii



import unittest

from xdrone.command_generators import generate_commands
from xdrone.compiler.compiler_utils.drones import NullDrone, Drone
from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.symbol_table import SymbolTable
from xdrone.compiler.compiler_utils.type import Type
from xdrone.shared.compile_error import CompileError
from xdrone.shared.drone_config import DefaultDroneConfig


class DeclareTest(unittest.TestCase):
    def test_declare_int_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { int a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 0, ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_decimal_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { decimal a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.decimal(), 0, ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_string_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { string a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.string(), "", ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_boolean_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { boolean a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), False, ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_vector_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { vector a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.vector(), [0, 0, 0], ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_drone_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { drone a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.drone(), NullDrone(), ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_list_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { list[int] a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.int()), [], ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_nested_list_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { list[list[int]] a; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.list_of(Type.int())), [], ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_with_different_variable_name_should_change_symbol_table(self):
        for name in ["a", "A", "_a", "_1", "_A", "abc", "Abc", "a12", "aA1", "_aA1"]:
            actual = SymbolTable()
            generate_commands("""
                main () {{ int {}; }}
                """.format(name), symbol_table=actual)
            expected = SymbolTable()
            expected.store(name, Expression(Type.int(), 0, ident=name))
            self.assertEqual(expected, actual)

    def test_repeated_declare_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 int a;
                 int a;
                }
                """)

        self.assertTrue("Identifier a already declared" in str(context.exception))

    def test_repeated_declare_different_type_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 int a;
                 decimal a;
                }
                """)

        self.assertTrue("Identifier a already declared" in str(context.exception))

    def test_repeated_declare_drone_constant_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                     {} DRONE;
                    }}
                    """.format(type.type_name), drone_config_map={"DRONE": DefaultDroneConfig()})

            self.assertTrue("Identifier DRONE already declared" in str(context.exception))


class AssignIdentTest(unittest.TestCase):
    def test_assign_ident_int_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             int a;
             a <- -1;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), -1, ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_decimal_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             decimal a;
             a <- -1.5e10;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.decimal(), -1.5e10, ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_string_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             string a;
             a <- "1";
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.string(), "1", ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_boolean_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             boolean a;
             a <- true;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_vector_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             vector a;
             a <- (1.0, -1.0, +1.0);
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.vector(), [1, -1, 1], ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_drone_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             drone a;
             a <- DRONE1;
            }
            """, drone_config_map={"DRONE1": DefaultDroneConfig()}, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.drone(), Drone("DRONE1", DefaultDroneConfig()), ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_list_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             list[int] a;
             a <- [1, -1, +1];
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.int()), [1, -1, 1], ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_nested_list_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             list[list[int]] a;
             a <- [[1], [-1], [+1]];
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.list_of(Type.int())), [[1], [-1], [1]], ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_ident_empty_list_should_update_symbol_table(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                 list[{}] a;
                 a <- [];
                }}
                """.format(type.type_name), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(Type.list_of(type), [], ident="a"))
            self.assertEqual(expected, actual)

    def test_assign_not_declared_variable_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 a <- true;
                }
            """)

        self.assertTrue("Identifier a has not been declared" in str(context.exception))

    def test_assign_drone_constant_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 DRONE1 <- DRONE1;
                }
            """, drone_config_map={"DRONE1": DefaultDroneConfig()})

        self.assertTrue("Identifier DRONE1 is a drone constant, cannot be assigned" in str(context.exception))

    def test_declare_and_then_assign_with_different_type_should_give_error(self):
        types = ["int", "decimal", "string", "boolean", "vector", "drone", "list[int]", "list[decimal]",
                 "list[list[int]]"]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                         {} a;
                         {} b;
                         a <- b;
                        }}
                    """.format(t1, t2))

                self.assertTrue("Identifier a has been declared as {}, but assigned as {}"
                                .format(t1, t2) in str(context.exception))

    def test_declare_and_then_assign_with_same_type_should_success(self):
        types = ["int", "decimal", "string", "boolean", "vector", "drone", "list[int]", "list[decimal]",
                 "list[list[int]]"]
        for type in types:
            generate_commands("""
                main () {{
                 {} a;
                 {} b;
                 a <- b;
                }}
            """.format(type, type))


class AssignVectorElemTest(unittest.TestCase):

    def test_assign_vector_elem_to_temp_vector_should_not_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             (0.0, 0.0, 0.0).x <- 1.0;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        self.assertEqual(expected, actual)

    def test_assign_vector_elem_to_variable_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             vector a;
             vector b;
             a.x <- 1;
             a.y <- 2;
             a.z <- -2;
             b.x <- 1.0;
             b.y <- 2.0;
             b.z <- -2.0;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.vector(), [1.0, 2.0, -2.0], ident="a"))
        expected.store("b", Expression(Type.vector(), [1.0, 2.0, -2.0], ident="b"))
        self.assertEqual(expected, actual)

    def test_assign_vector_elem_not_declared_variable_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
            main () {
             a.x <- 1.0;
            }
            """)

        self.assertTrue("Identifier a has not been declared" in str(context.exception))

    def test_declare_and_then_assign_vector_elem_with_different_type_should_give_error(self):
        for type in [Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                     Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]:
            for index in ["x", "y", "z"]:
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                         vector a;
                         {} b;
                         a.{} <- b;
                        }}
                    """.format(type.type_name, index))

                self.assertTrue("Assigned value {} should have type int or decimal, but is {}"
                                .format(Expression(type, type.default_value, ident="b"), type.type_name)
                                in str(context.exception))


class AssignListElemTest(unittest.TestCase):

    def test_assign_list_elem_to_temp_vector_should_not_change_symbol_table(self):
        for code in ["[0.0][0] <- 1.0", "[0][0] <- 1", "[[\"a\"]][0] <- [\"b\"]", "[0.0, 1.0][0] <- 1.0"]:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                 {};
                }}
                """.format(code), symbol_table=actual)
            expected = SymbolTable()
            self.assertEqual(expected, actual)

    def test_assign_list_elem_to_variable_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             list[int] a <- [0, 1, 2];
             a[0] <- 1;
             a[2] <- a[0];
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.int()), [1, 1, 1], ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_list_elem_to_variable_nested_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             list[list[int]] a <- [[0, 1], [2, 3]];
             a[0] <- [4, 5];
             a[1][0] <- 6;
             a[1][1] <- 7;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.list_of(Type.int())), [[4, 5], [6, 7]], ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_list_elem_with_vector_should_update_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             list[vector] a <- [(1.0, 2.0, 3.0), (4.0, 5.0, 6.0)];
             a[0] <- (7.0, 8.0, 9.0);
             a[1].x <- 10.0;
             a[1].y <- 11.0;
             a[1].z <- 12.0;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.vector()), [[7, 8, 9], [10, 11, 12]], ident="a"))
        self.assertEqual(expected, actual)

    def test_assign_list_elem_to_variable_out_of_bound_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 list[int] a <- [];
                 a[0] <- 1;
                }
                """)

        self.assertTrue("List {} has length 0, but has been assessed with out-of-range index 0"
                        .format(Expression(Type.list_of(Type.int()), [], ident="a"))
                        in str(context.exception))

    def test_assess_list_elem_out_of_bound_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 list[int] a <- [1];
                 a[0] <- a[1];
                }
                """)

        self.assertTrue("List {} has length 1, but has been assessed with out-of-range index 1"
                        .format(Expression(Type.list_of(Type.int()), [1], ident="a"))
                        in str(context.exception))

    def test_assess_list_elem_nested_out_of_bound_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 list[list[list[int]]] a <- [[[1], [2]]];
                 a[0][2] <- [1];
                }
                """)
        self.assertTrue("List {} has length 2, but has been assessed with out-of-range index 2"
                        .format(Expression(Type.list_of(Type.list_of(Type.int())), [[1], [2]], ident="a[0]"))
                        in str(context.exception))

        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 list[list[list[int]]] a <- [[[1], [2]]];
                 a[0][1][1] <- 1;
                }
                """)
        self.assertTrue("List {} has length 1, but has been assessed with out-of-range index 1"
                        .format(Expression(Type.list_of(Type.int()), [2], ident="a[0][1]"))
                        in str(context.exception))

    def test_assign_list_elem_not_declared_variable_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
            main () {
             a[0] <- 1;
            }
            """)

        self.assertTrue("Identifier a has not been declared" in str(context.exception))

    def test_declare_and_assign_list_elem_with_different_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                    main () {{
                     {} a;
                     list[{}] b <- [a];
                    }}
                    """.format(t1, t2))

                self.assertTrue("Identifier b has been declared as list[{}], but assigned as list[{}]"
                                .format(t2, t1) in str(context.exception))

    def test_assign_list_elem_with_different_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                    main () {{
                     {} a;
                     {} b;
                     list[{}] c <- [a];
                     c[0] <- b;
                    }}
                    """.format(t1, t2, t1))

                self.assertTrue("Assigned value {} should have type {}, but is {}"
                                .format(Expression(t2, t2.default_value, ident="b"), t1.type_name, t2.type_name)
                                in str(context.exception))


class CombinedDeclareAssignTest(unittest.TestCase):

    def test_declare_and_assign_int_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { int a <- 1; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 1, ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_decimal_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { decimal a <- 1.0; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.decimal(), 1.0, ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_string_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { string a <- "\0a"; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.string(), "\0a", ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_boolean_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { boolean a <- true; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.boolean(), True, ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_vector_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { vector a <- (1.0, 2.0, -3.0); }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.vector(), [1, 2, -3], ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_drone_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { drone a <- DRONE1; }
            """, drone_config_map={"DRONE1": DefaultDroneConfig()}, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.drone(), Drone("DRONE1", DefaultDroneConfig()), ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_list_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { list[int] a <- [1, 2, -3]; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.int()), [1, 2, -3], ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_nested_list_should_change_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () { list[list[int]] a <- [[1], [2]]; }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.list_of(Type.list_of(Type.int())), [[1], [2]], ident="a"))
        self.assertEqual(expected, actual)

    def test_declare_and_assign_ident_empty_list_should_update_symbol_table(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            actual = SymbolTable()
            generate_commands("""
                main () {{
                 list[{}] a <- [];
                }}
                """.format(type.type_name), symbol_table=actual)
            expected = SymbolTable()
            expected.store("a", Expression(Type.list_of(type), [], ident="a"))
            self.assertEqual(expected, actual)

    def test_repeated_declare_and_assign_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 int a;
                 int a <- 0;
                }
                """)

        self.assertTrue("Identifier a already declared" in str(context.exception))

    def test_repeated_declare_and_assign_drone_constant_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                     {} a;
                     {} DRONE <- a;
                    }}
                    """.format(type.type_name, type.type_name), drone_config_map={"DRONE": DefaultDroneConfig()})

            self.assertTrue("Identifier DRONE already declared" in str(context.exception))

    def test_declare_and_assign_with_different_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.list_of(Type.int()))]
        for t1 in types:
            for t2 in types:
                if t1 == t2:
                    continue
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                    main () {{
                     {} a;
                     {} b <- a;
                    }}
                    """.format(t1, t2))

                self.assertTrue("Identifier b has been declared as {}, but assigned as {}"
                                .format(t2.type_name, t1.type_name) in str(context.exception))


class DelTest(unittest.TestCase):
    def test_del_should_delete_symbol_table(self):
        actual = SymbolTable()
        generate_commands("""
            main () {
             int a;
             del a;
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        self.assertEqual(expected, actual)

    def test_del_not_declared_variable_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
            main () {
             del a;
            }
            """)

        self.assertTrue("Identifier a has not been declared" in str(context.exception))

    def test_del_drone_constant_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                 del DRONE1;
                }
            """, drone_config_map={"DRONE1": DefaultDroneConfig()})

        self.assertTrue("Identifier DRONE1 is a drone constant, cannot be deleted" in str(context.exception))



import unittest
from unittest.mock import Mock, call

from xdrone.safety_checker.boundary_checker import BoundaryChecker
from xdrone.shared.boundary_config import BoundaryConfig
from xdrone.shared.command import Command, SingleDroneCommand, ParallelDroneCommands
from xdrone.shared.drone_config import DroneConfig
from xdrone.shared.safety_check_error import SafetyCheckError
from xdrone.shared.state import State
from xdrone.state_updaters.state_updater import StateUpdater


class BoundaryCheckerTest(unittest.TestCase):
    def setUp(self) -> None:
        self.boundary_config = BoundaryConfig(max_seconds=10, max_x_meters=5, max_y_meters=5, max_z_meters=5,
                                              min_x_meters=-5, min_y_meters=-5, min_z_meters=-5)
        self.state_updater_map = {"DRONE1": StateUpdater(DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                                     rotate_speed_dps=90, takeoff_height_meters=1)),
                                  "DRONE2": StateUpdater(DroneConfig(init_position=(1, 0, 0), speed_mps=2,
                                                                     rotate_speed_dps=180, takeoff_height_meters=2))}

    def test_check_should_call_update_states_and_check_drone_commands_with_correct_parameters(self):
        boundary_checker = BoundaryChecker(self.boundary_config)
        boundary_checker._update_states_and_check_drone_commands = Mock()
        drone_state_map = {"DRONE1": State(), "DRONE2": State(x_meters=1)}
        drones_involved = {"DRONE1", "DRONE2"}
        boundary_checker.check([], self.state_updater_map)
        boundary_checker._update_states_and_check_drone_commands.assert_called_once_with([], self.state_updater_map,
                                                                                         drone_state_map,
                                                                                         drones_involved)

    def test_check_in_the_end_not_land_should_give_error(self):
        with self.assertRaises(SafetyCheckError) as context:
            drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff())]
            BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)
        self.assertTrue("Drone 'DRONE1' did not land in the end" in str(context.exception))

    def test_check_takeoff_land_takeoff_land_should_not_give_error(self):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land()),
                          SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land()), ]
        BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)

    def test_check_single_drone_command_wait_when_taken_off_should_not_give_error(self):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.wait(1)),
                          SingleDroneCommand("DRONE1", Command.land())]
        BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)

    def test_check_single_drone_command_wait_when_not_taken_off_should_not_give_error(self):
        drone_commands = [SingleDroneCommand("DRONE1", Command.wait(1))]
        BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)

    def test_check_single_drone_command_takeoff_when_taken_off_should_give_error(self):
        with self.assertRaises(SafetyCheckError) as context:
            drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                              SingleDroneCommand("DRONE1", Command.takeoff())]
            BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)
        self.assertTrue("'takeoff' command used when drone 'DRONE1' has already been taken off"
                        in str(context.exception))

    def test_check_single_drone_command_other_command_when_not_taken_off_should_give_error(self):
        commands = [Command.land(), Command.up(1), Command.down(1), Command.left(1), Command.right(1),
                    Command.forward(1), Command.backward(1), Command.rotate_left(90), Command.rotate_right(90)]
        for command in commands:
            with self.assertRaises(SafetyCheckError) as context:
                drone_commands = [SingleDroneCommand("DRONE1", command)]
                BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)
            self.assertTrue("'{}' command used when drone 'DRONE1' has not been taken off"
                            .format(command.opcode)
                            in str(context.exception))

    def test_check_single_drone_command_should_update_state_correctly(self):
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff())]
        drone_state_map = {"DRONE1": State(), "DRONE2": State(x_meters=1)}
        drones_involved = {"DRONE1", "DRONE2"}
        BoundaryChecker(self.boundary_config)._update_states_and_check_drone_commands(drone_commands,
                                                                                      self.state_updater_map,
                                                                                      drone_state_map,
                                                                                      drones_involved)
        expected = {"DRONE1": State(has_taken_off=True, time_used_seconds=1, z_meters=1),
                    "DRONE2": State(time_used_seconds=1, x_meters=1)}
        self.assertEqual(expected, drone_state_map)

    def test_check_single_drone_command_should_check_state(self):
        boundary_checker = BoundaryChecker(self.boundary_config)
        boundary_checker.boundary_config.check_state = Mock()
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land())]
        BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)
        calls = [call("DRONE1", State(has_taken_off=True, time_used_seconds=1, z_meters=1)),
                 call("DRONE2", State(time_used_seconds=1, x_meters=1)),
                 call("DRONE1", State(time_used_seconds=2)),
                 call("DRONE2", State(time_used_seconds=2, x_meters=1))]
        boundary_checker.boundary_config.check_state.assert_has_calls(calls)

    def test_check_single_drone_command_should_check_state_and_catch_error(self):
        boundary_checker = BoundaryChecker(self.boundary_config)
        boundary_checker.boundary_config.check_state = Mock(side_effect=SafetyCheckError)
        drone_commands = [SingleDroneCommand("DRONE1", Command.takeoff()),
                          SingleDroneCommand("DRONE1", Command.land())]
        with self.assertRaises(SafetyCheckError) as context:
            BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)

    def test_check_parallel_drone_commands_should_wait_for_slower_branch(self):
        drone_commands = [ParallelDroneCommands([
            [SingleDroneCommand("DRONE1", Command.takeoff())],
            [SingleDroneCommand("DRONE2", Command.wait(5))]
        ])]
        drone_state_map = {"DRONE1": State(), "DRONE2": State(x_meters=1)}
        drones_involved = {"DRONE1", "DRONE2"}
        BoundaryChecker(self.boundary_config)._update_states_and_check_drone_commands(drone_commands,
                                                                                      self.state_updater_map,
                                                                                      drone_state_map,
                                                                                      drones_involved)
        expected = {"DRONE1": State(has_taken_off=True, time_used_seconds=5, z_meters=1),
                    "DRONE2": State(time_used_seconds=5, x_meters=1)}
        self.assertEqual(expected, drone_state_map)

    def test_check_parallel_drone_commands_should_let_drone_not_involved_wait(self):
        drone_commands = [ParallelDroneCommands([
            [],
            [SingleDroneCommand("DRONE2", Command.wait(5))]
        ])]
        drone_state_map = {"DRONE1": State(), "DRONE2": State(x_meters=1)}
        drones_involved = {"DRONE1", "DRONE2"}
        BoundaryChecker(self.boundary_config)._update_states_and_check_drone_commands(drone_commands,
                                                                                      self.state_updater_map,
                                                                                      drone_state_map,
                                                                                      drones_involved)
        expected = {"DRONE1": State(time_used_seconds=5),
                    "DRONE2": State(time_used_seconds=5, x_meters=1)}
        self.assertEqual(expected, drone_state_map)

    def test_check_parallel_drone_commands_nested_should_update_states_correctly(self):
        drone_commands = \
            [
                ParallelDroneCommands([
                    [SingleDroneCommand("DRONE1", Command.wait(5))],
                    [SingleDroneCommand("DRONE2", Command.wait(3)),
                     ParallelDroneCommands([
                         [SingleDroneCommand("DRONE3", Command.wait(4))],
                         [SingleDroneCommand("DRONE4", Command.wait(1))]
                     ])]
                ]),
                SingleDroneCommand("DRONE1", Command.wait(1))
            ]
        drones_involved = {"DRONE1", "DRONE2", "DRONE3", "DRONE4", "DRONE5"}
        state_updaters = {name: StateUpdater(DroneConfig((1, 0, 0), 2, 180, 2)) for name in drones_involved}
        drone_state_map = {name: State() for name in drones_involved}
        BoundaryChecker(self.boundary_config)._update_states_and_check_drone_commands(drone_commands,
                                                                                      state_updaters,
                                                                                      drone_state_map,
                                                                                      drones_involved)
        expected = {name: State(time_used_seconds=8) for name in drones_involved}
        self.assertEqual(expected, drone_state_map)

    def test_check_parallel_drone_commands_should_check_state(self):
        boundary_checker = BoundaryChecker(self.boundary_config)
        boundary_checker.boundary_config.check_state = Mock()
        drone_commands = [ParallelDroneCommands([
            [SingleDroneCommand("DRONE1", Command.takeoff()), SingleDroneCommand("DRONE1", Command.land())],
            [SingleDroneCommand("DRONE2", Command.takeoff()), SingleDroneCommand("DRONE2", Command.land())]
        ])]
        BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)
        last_two_calls = [call("DRONE1", State(time_used_seconds=2)),
                          call("DRONE2", State(time_used_seconds=2, x_meters=1))]
        boundary_checker.boundary_config.check_state.assert_has_calls(last_two_calls)

    def test_check_parallel_drone_commands_should_check_state_and_catch_error(self):
        boundary_checker = BoundaryChecker(self.boundary_config)
        boundary_checker.boundary_config.check_state = Mock(side_effect=SafetyCheckError)
        drone_commands = [ParallelDroneCommands([
            [],
            []
        ])]
        with self.assertRaises(SafetyCheckError) as context:
            BoundaryChecker(self.boundary_config).check(drone_commands, self.state_updater_map)



import unittest

from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.symbol_table import SymbolTable
from xdrone.compiler.compiler_utils.type import Type


class TestSymbolTable(unittest.TestCase):
    def test_contains(self):
        st = SymbolTable()
        self.assertFalse("a" in st)
        st.store("a", Expression(Type.int(), 0))
        self.assertTrue("a" in st)

    def test_eq(self):
        st1 = SymbolTable()
        st2 = SymbolTable()
        self.assertTrue(st1 == st2)
        st1.store("a", Expression(Type.int(), 0))
        self.assertFalse(st1 == st2)
        st2.store("a", Expression(Type.int(), 0))
        self.assertTrue(st1 == st2)

        self.assertNotEqual(SymbolTable(), None)

    def test_str(self):
        st = SymbolTable()
        st.store("a", Expression(Type.int(), 0))
        st.store("b", Expression(Type.list_of(Type.boolean()), [True]))
        expected = "SymbolTable: {\n" + \
                   "  a -> Expression: { type: int, value: 0, ident: None }\n" + \
                   "  b -> Expression: { type: list[boolean], value: [True], ident: None }\n" + \
                   "}"
        self.assertEqual(expected, str(st))

    def test_store(self):
        st = SymbolTable()
        st.store("a", Expression(Type.int(), 1))
        self.assertTrue("a" in st)
        self.assertEqual(Type.int(), st.get_expression("a").type)
        self.assertEqual(1, st.get_expression("a").value)
        self.assertRaises(AssertionError, st.store, "a", Expression(Type.int(), 1))

    def test_update(self):
        st = SymbolTable()
        self.assertRaises(AssertionError, st.update, "a", 1)
        st.store("a", Expression(Type.int(), 1))
        st.update("a", 2)
        self.assertEqual(Type.int(), st.get_expression("a").type)
        self.assertEqual(2, st.get_expression("a").value)

    def test_delete(self):
        st = SymbolTable()
        st.store("a", Expression(Type.int(), 1))
        self.assertTrue("a" in st)
        st.delete("a")
        self.assertFalse("a" in st)

    def test_get_expression(self):
        st = SymbolTable()
        self.assertRaises(AssertionError, st.get_expression, "a")
        st.store("a", Expression(Type.list_of(Type.int()), [1]))
        self.assertEqual([1], st.get_expression("a").value)
        self.assertEqual(1, st.get_expression("a").value[0])
        self.assertEqual(Type.list_of(Type.int()), st.get_expression("a").type)



from __future__ import annotations

import copy
from typing import Union

from xdrone.compiler.compiler_utils.drones import Drone, NullDrone


class Type:

    def __init__(self, type_name: str, default_value: Union[int, float, str, bool, Drone, list]):
        self._type_name = type_name
        self._default_value = default_value

    @property
    def type_name(self) -> str:
        return copy.deepcopy(self._type_name)

    @property
    def default_value(self) -> Union[int, float, str, bool, Drone, list]:
        return copy.deepcopy(self._default_value)

    @staticmethod
    def int() -> Type:
        return Type("int", 0)

    @staticmethod
    def decimal() -> Type:
        return Type("decimal", 0.0)

    @staticmethod
    def string() -> Type:
        return Type("string", "")

    @staticmethod
    def boolean() -> Type:
        return Type("boolean", False)

    @staticmethod
    def vector() -> Type:
        return Type("vector", [0.0, 0.0, 0.0])

    @staticmethod
    def drone() -> Type:
        return Type("drone", NullDrone())

    @staticmethod
    def list_of(elem_type: Type) -> ListType:
        return ListType(elem_type)

    @staticmethod
    def empty_list() -> EmptyList:
        return EmptyList()

    def __str__(self):
        return self.type_name

    def __eq__(self, other):
        if isinstance(other, Type):
            return other.type_name == self.type_name and other.default_value == self.default_value
        return False


class ListType(Type):
    def __init__(self, elem_type: Type):
        type_name = "list[" + elem_type.type_name + "]"
        super().__init__(type_name, [])
        self._elem_type = elem_type

    @property
    def elem_type(self) -> Type:
        return copy.deepcopy(self._elem_type)

    def __eq__(self, other):
        if isinstance(other, ListType):
            return other.type_name == self.type_name and other.default_value == self.default_value \
                   and other.elem_type == self.elem_type
        return False


class EmptyList(ListType):
    def __init__(self):
        super().__init__(Type("all", 0))
        self._type_name = "list[]"

    def __eq__(self, other):
        return isinstance(other, ListType)



import unittest

from xdrone.shared.state import State


class StateTest(unittest.TestCase):
    def test_init_state(self):
        state = State()
        self.assertEqual(False, state.has_taken_off)
        self.assertEqual(0, state.time_used_seconds)
        self.assertEqual(0, state.x_meters)
        self.assertEqual(0, state.y_meters)
        self.assertEqual(0, state.z_meters)
        self.assertEqual(0, state.orientation_degrees)

    def test_property(self):
        state = State(has_taken_off=True, time_used_seconds=1,
                      x_meters=1, y_meters=1, z_meters=1, orientation_degrees=1)
        self.assertEqual(True, state.has_taken_off)
        self.assertEqual(1, state.time_used_seconds)
        self.assertEqual(1, state.x_meters)
        self.assertEqual(1, state.y_meters)
        self.assertEqual(1, state.z_meters)
        self.assertEqual(1, state.orientation_degrees)

    def test_copy_and_set_has_taken_off(self):
        state = State()
        actual = state.copy_and_set_has_taken_off(True)
        expected = State(has_taken_off=True, time_used_seconds=0,
                         x_meters=0, y_meters=0, z_meters=0, orientation_degrees=0)
        self.assertEqual(expected, actual)

    def test_copy_and_set_time_used_seconds(self):
        state = State()
        actual = state.copy_and_set_time_used_seconds(1)
        expected = State(has_taken_off=False, time_used_seconds=1,
                         x_meters=0, y_meters=0, z_meters=0, orientation_degrees=0)
        self.assertEqual(expected, actual)

    def test_copy_and_set_x_meters(self):
        state = State()
        actual = state.copy_and_set_x_meters(1)
        expected = State(has_taken_off=False, time_used_seconds=0,
                         x_meters=1, y_meters=0, z_meters=0, orientation_degrees=0)
        self.assertEqual(expected, actual)

    def test_copy_and_set_y_meters(self):
        state = State()
        actual = state.copy_and_set_y_meters(1)
        expected = State(has_taken_off=False, time_used_seconds=0,
                         x_meters=0, y_meters=1, z_meters=0, orientation_degrees=0)
        self.assertEqual(expected, actual)

    def test_copy_and_set_z_meters(self):
        state = State()
        actual = state.copy_and_set_z_meters(1)
        expected = State(has_taken_off=False, time_used_seconds=0,
                         x_meters=0, y_meters=0, z_meters=1, orientation_degrees=0)
        self.assertEqual(expected, actual)

    def test_copy_and_set_orientation_degrees(self):
        state = State()
        actual = state.copy_and_set_orientation_degrees(1)
        expected = State(has_taken_off=False, time_used_seconds=0,
                         x_meters=0, y_meters=0, z_meters=0, orientation_degrees=1)
        self.assertEqual(expected, actual)

    def test_str(self):
        state = State()
        self.assertEqual("State: { has_taken_off: False, time_used_seconds: 0, " +
                         "x_meters: 0, y_meters: 0, z_meters: 0, orientation_degrees: 0 }",
                         str(state))

    def test_eq(self):
        self.assertEqual(State(), State())
        self.assertNotEqual(State(time_used_seconds=False), State(has_taken_off=True))
        self.assertNotEqual(State(time_used_seconds=0), State(time_used_seconds=1))
        self.assertNotEqual(State(x_meters=0), State(x_meters=1))
        self.assertNotEqual(State(y_meters=0), State(y_meters=1))
        self.assertNotEqual(State(z_meters=0), State(z_meters=1))
        self.assertNotEqual(State(orientation_degrees=0), State(orientation_degrees=1))
        self.assertNotEqual(None, State())



from setuptools import setup, find_packages

with open("README.md") as readme:
    long_description = readme.read()

# This call to setup() does all the work
setup(
    name="xdrone",
    version="0.2.8",
    description="xDrone Language Server",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/xDrone-DSL/xDroneLanguageServer",
    author="Kai Zhu",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=["click", "antlr4-python3-runtime", "netifaces", "netaddr", "scipy", "matplotlib", "tabulate"],
    entry_points={
        "console_scripts": [
            "xdrone=cmdline.xdrone:main",
        ]
    },
)



#
# Quick and basic code for running a standalone "emergency stop" in case of any problems
#
#
#

import socket
import netaddr
import netifaces
import time


#
# FUNCTION DEFINITIONS
#

def _get_subnets():
    """ Get the local subnet and server IP address """
    subnets = []
    addr_list = []
    ifaces = netifaces.interfaces()
    for this_iface in ifaces:
        addrs = netifaces.ifaddresses(this_iface)

        if socket.AF_INET not in addrs:
            continue

        # Get IPv4 info
        ip_info = addrs[socket.AF_INET][0]
        address = ip_info['addr']
        netmask = ip_info['netmask']

        # Avoid searching when on very large subnets
        if netmask != '255.255.255.0':
            continue

        # Create IP object and get the network details
        # Note CIDR is a networking term, describing the IP/subnet address format
        cidr = netaddr.IPNetwork('%s/%s' % (address, netmask))
        network = cidr.network
        subnets.append((network, netmask))
        addr_list.append(address)
    return subnets, addr_list


def send_command(command, possible_addr, control_socket, control_port):
    # Send the command to each Tello on each possible_addr
    for ip in possible_addr:
        try:
            print('Sending %s command to drone at %s' % (command, ip))
            control_socket.sendto(command.encode(), (ip, control_port))
        except OSError as oserror:
            print(oserror)
            print('ERROR! Socket failed - terminating!')
            exit()
        time.sleep(0.01)


def initialise(first_ip, last_ip, control_port, possible_addr):

    # Create socket for communication with Tello
    control_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # socket for sending cmd
    control_socket.bind(('', control_port))

    # Get network addresses to search
    subnets, address = _get_subnets()

    # Create a list of possible IP addresses to search
    for subnet, netmask in subnets:
        for ip in netaddr.IPNetwork('%s/%s' % (subnet, netmask)):
            if not (first_ip <= int(str(ip).split('.')[3]) <= last_ip):
                continue
            # Don't add the server's address to the list
            if str(ip) in address:
                continue
            possible_addr.append(str(ip))

    return possible_addr, control_socket


#
# MAIN SCRIPT
#

print('Emergency Stop Application Started!')
first_ip = 51
last_ip = 54
control_port = 8889
control_socket = None
possible_addr = []

while True:
    # Do nothing until we've got a command
    command = input('Emergency Stop?  ' ' or L = Auto-Land  |  S = Stop  |  E = Emergency Cut-Out  |  Q = Quit: ')

    # If not already initalised, initialise the network connection via a UDP socket
    if not control_socket:
        possible_addr, control_socket = initialise(first_ip, last_ip, control_port, possible_addr)
        print('Connection initialised!')

    if command.upper() == ' ':
        send_command('land', possible_addr, control_socket, control_port)
    elif command.upper() == 'L':
        send_command('land', possible_addr, control_socket, control_port)
    elif command.upper() == 'S':
        send_command('stop', possible_addr, control_socket, control_port)
    elif command.upper() == 'E':
        send_command('emergency', possible_addr, control_socket, control_port)
    elif command.upper() == 'Q':
        print('Q(uit) command received - exiting!')
        exit()
    else:
        print('Invalid command - enter space (to land), L(and), S(top), E(mergency), or Q(uit)')

    time.sleep(0.1)



import unittest

from xdrone.command_generators import generate_commands
from xdrone.compiler.compiler_utils.expressions import Expression
from xdrone.compiler.compiler_utils.symbol_table import SymbolTable
from xdrone.compiler.compiler_utils.type import Type
from xdrone.shared.command import Command, SingleDroneCommand
from xdrone.shared.compile_error import CompileError


class IfTest(unittest.TestCase):
    def test_if_true_without_else_should_run_correct_commands(self):
        actual_st = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              if true {
                int a <- 1;
                forward(a);
              }
              land();
            }
            """, symbol_table=actual_st)
        expected_st = SymbolTable()
        self.assertEqual(expected_st, actual_st)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(1)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_if_true_with_else_should_run_correct_commands(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              if true {
                int a <- 1;
                forward(a);
              } else {
                int a <- 2;
                forward(a);
              }
              land();
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        self.assertEqual(expected, actual)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(1)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_if_false_without_else_should_do_nothing(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              if false {
                int a <- 1;
              }
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        self.assertEqual(expected, actual)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)

    def test_if_false_with_else_should_run_correct_commands(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              if false {
                int a <- 1;
                forward(a);
              } else {
                int a <- 2;
                int b <- 3;
                forward(b);
              }
              land();
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        self.assertEqual(expected, actual)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.forward(3)),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_if_with_error_commands_not_entering_should_not_give_error(self):
        actual_commands = generate_commands("""
            main () {
              takeoff();
              if false {
                int a <- "error";
                forward(1);
              }
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_if_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      if a {{
                        int b <- 1;
                      }} else {{
                        int b <- 2;
                      }}
                    }}
                    """.format(type))

            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_if_should_keep_updates_but_discard_new_variables(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              int a <- 1;
              if true {
                int b <- 1;
                a <- 2;
              }
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 2, ident="a"))
        self.assertEqual(expected, actual)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)

    def test_if_with_else_should_keep_updates_but_discard_new_variables(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              int a <- 1;
              if false {
              } else {
                int b <- 1;
                a <- 2;
              }
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 2, ident="a"))
        self.assertEqual(expected, actual)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)


class WhileTest(unittest.TestCase):
    def test_while_should_run_correct_commands(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              int a <- 1;
              while a < 5 {
                a <- a + 1;
                forward(a);
              }
              land();
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 5, ident="a"))
        self.assertEqual(expected, actual)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff())] + \
                            [SingleDroneCommand("DEFAULT", Command.forward(i)) for i in [2, 3, 4, 5]] + \
                            [SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_while_false_should_not_enter_loop(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              int a <- 10;
              while a < 5 {
                a <- a + 1;
                forward(a);
              }
              land();
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 10, ident="a"))
        self.assertEqual(expected, actual)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_while_with_error_command_not_entering_should_not_give_error(self):
        actual_commands = generate_commands("""
            main () {
              while false {
                int a <- "error";
              }
            }
            """)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)

    def test_while_wrong_type_should_give_error(self):
        types = [Type.int(), Type.decimal(), Type.string(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      int b <- 1;
                      while a {{
                        b <- b + 1;
                      }}
                    }}
                    """.format(type))

            self.assertTrue("Expression {} should have type boolean, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_while_should_keep_updates_but_discard_new_variables(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              int a <- 0;
              while a < 5 {
                int b <- 1;
                a <- a + 1;
              }
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 5, ident="a"))
        self.assertEqual(expected, actual)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)


class ForTest(unittest.TestCase):
    def test_for_without_step_should_run_correct_commands_and_update_symbol_table(self):
        actual_st = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              int i;
              int a <- 0;
              for i from 0 to 5 {
                a <- a + 1;
                forward(i);
              }
              land();
            }
            """, symbol_table=actual_st)
        expected_st = SymbolTable()
        expected_st.store("i", Expression(Type.int(), 5, ident="i"))
        expected_st.store("a", Expression(Type.int(), 6, ident="a"))
        self.assertEqual(expected_st, actual_st)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff())] + \
                            [SingleDroneCommand("DEFAULT", Command.forward(i)) for i in [0, 1, 2, 3, 4, 5]] + \
                            [SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_for_with_step_should_run_correct_commands_and_update_symbol_table(self):
        actual_st = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              int i;
              int a <- 0;
              for i from 0 to 9 step 2 {
                a <- a + 1;
                forward(i);
              }
              land();
            }
            """, symbol_table=actual_st)
        expected_st = SymbolTable()
        expected_st.store("i", Expression(Type.int(), 8, ident="i"))
        expected_st.store("a", Expression(Type.int(), 5, ident="a"))
        self.assertEqual(expected_st, actual_st)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff())] + \
                            [SingleDroneCommand("DEFAULT", Command.forward(i)) for i in [0, 2, 4, 6, 8]] + \
                            [SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_for_not_entering_should_not_update_symbol_table(self):
        actual_st = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              takeoff();
              int i;
              int a <- 0;
              for i from 10 to 5 {
                a <- a + 1;
                forward(i);
              }
              land();
            }
            """, symbol_table=actual_st)
        expected_st = SymbolTable()
        expected_st.store("i", Expression(Type.int(), 0, ident="i"))
        expected_st.store("a", Expression(Type.int(), 0, ident="a"))
        self.assertEqual(expected_st, actual_st)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff()),
                             SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_for_with_error_commands_not_entering_should_not_give_error(self):
        actual_commands = generate_commands("""
            main () {
              int i;
              for i from 10 to 5 {
                int a <- "error";
              }
            }
            """)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)

    def test_for_with_not_declared_ident_should_give_error(self):
        with self.assertRaises(CompileError) as context:
            generate_commands("""
                main () {
                  for i from 0 to 9 step 2 {
                  }
                }
                """)

        self.assertTrue("Identifier i has not been declared" in str(context.exception))

    def test_for_with_wrong_type_ident_should_give_error(self):
        types = [Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} i;
                      for i from 0 to 9 step 2 {{
                      }}
                    }}
                    """.format(type))

            self.assertTrue("Identifier i has been declared as {}, but assigned as int"
                            .format(type)
                            in str(context.exception))

    def test_for_with_wrong_type_from_expr_should_give_error(self):
        types = [Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            for with_step in ["", "step 2"]:
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          int i;
                          {} a;
                          for i from a to 9 {} {{
                          }}
                        }}
                        """.format(type, with_step))

                self.assertTrue("Expression {} should have type int, but is {}"
                                .format(Expression(type, type.default_value, ident="a"), type.type_name)
                                in str(context.exception))

    def test_for_with_wrong_type_to_expr_should_give_error(self):
        types = [Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            for with_step in ["", "step 2"]:
                with self.assertRaises(CompileError) as context:
                    generate_commands("""
                        main () {{
                          int i;
                          {} a;
                          for i from 0 to a {} {{
                          }}
                        }}
                        """.format(type, with_step))

                self.assertTrue("Expression {} should have type int, but is {}"
                                .format(Expression(type, type.default_value, ident="a"), type.type_name)
                                in str(context.exception))

    def test_for_with_wrong_type_step_expr_should_give_error(self):
        types = [Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      int i;
                      {} a;
                      for i from 0 to 9 step a {{
                      }}
                    }}
                    """.format(type))

            self.assertTrue("Expression {} should have type int, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_for_should_keep_updates_but_discard_new_variables(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              int a <- 0;
              int i <- 1;
              for i from 1 to 5 {
                int b <- 1;
                a <- a + 1;
              }
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 5, ident="a"))
        expected.store("i", Expression(Type.int(), 5, ident="i"))
        self.assertEqual(expected, actual)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)


class RepeatTest(unittest.TestCase):
    def test_repeat_should_run_correct_commands(self):
        actual_commands = generate_commands("""
            main () {
              takeoff();
              repeat 4 times {
                forward(1);
              }
              land();
            }
            """)
        expected_commands = [SingleDroneCommand("DEFAULT", Command.takeoff())] + \
                            [SingleDroneCommand("DEFAULT", Command.forward(1)) for _ in range(4)] + \
                            [SingleDroneCommand("DEFAULT", Command.land())]
        self.assertEqual(expected_commands, actual_commands)

    def test_repeat_with_wrong_type_expr_should_give_error(self):
        types = [Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                 Type.list_of(Type.int()), Type.list_of(Type.decimal()), Type.list_of(Type.list_of(Type.int()))]
        for type in types:
            with self.assertRaises(CompileError) as context:
                generate_commands("""
                    main () {{
                      {} a;
                      repeat a times {{
                      }}
                    }}
                    """.format(type))

            self.assertTrue("Expression {} should have type int, but is {}"
                            .format(Expression(type, type.default_value, ident="a"), type.type_name)
                            in str(context.exception))

    def test_repeat_should_keep_updates_but_discard_new_variables(self):
        actual = SymbolTable()
        actual_commands = generate_commands("""
            main () {
              int a <- 0;
              repeat 5 times {
                int b <- 1;
                a <- a + 1;
              }
            }
            """, symbol_table=actual)
        expected = SymbolTable()
        expected.store("a", Expression(Type.int(), 5, ident="a"))
        self.assertEqual(expected, actual)
        expected_commands = []
        self.assertEqual(expected_commands, actual_commands)



import logging
import unittest

from xdrone.config_parsers.config_parser import ConfigParser
from xdrone.shared.boundary_config import BoundaryConfig
from xdrone.shared.collision_config import CollisionConfig, DefaultCollisionConfig
from xdrone.shared.drone_config import DroneConfig, DefaultDroneConfig


class ConfigParserTest(unittest.TestCase):
    def test_parse_if_provided_configs_should_parse_all_configs(self):
        config = """
            {
              "drones": [{
                "name": "DRONE1",
                "init_position": {"x": 1, "y": 2, "z": 3},
                "speed_mps": 2,
                "rotate_speed_dps": 180,
                "takeoff_height_meters": 2,
                "advanced": {
                  "variance_per_meter": 0.002
                }
              },{
                "name": "DRONE2",
                "init_position": {"x": 4, "y": 5, "z": 6},
                "speed_mps": 1,
                "rotate_speed_dps": 90,
                "takeoff_height_meters": 1,
                "advanced": {
                  "variance_per_meter": 0.001
                }
              }],
              "boundary_config": {
                "max_seconds": 100,
                "max_x_meters": 10,
                "max_y_meters": 20,
                "max_z_meters": 30,
                "min_x_meters": -10,
                "min_y_meters": -20,
                "min_z_meters": -30
              },
              "collision_config": {
                "collision_meters": 0.3,
                "time_interval_seconds": 0.5,
                "confidence_threshold": 0.9
              }
            }
            """
        actual_drone_config, actual_boundary_config, acutal_collision_config = ConfigParser.parse(config)
        expected_drone_configs = {"DRONE1": DroneConfig(init_position=(1, 2, 3), speed_mps=2,
                                                        rotate_speed_dps=180, takeoff_height_meters=2,
                                                        var_per_meter=0.002),
                                  "DRONE2": DroneConfig(init_position=(4, 5, 6), speed_mps=1,
                                                        rotate_speed_dps=90, takeoff_height_meters=1,
                                                        var_per_meter=0.001)}
        expected_boundary_config = BoundaryConfig(max_seconds=100, max_x_meters=10, max_y_meters=20, max_z_meters=30,
                                                  min_x_meters=-10, min_y_meters=-20, min_z_meters=-30)
        expected_collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5,
                                                    confidence_threshold=0.9)
        self.assertEqual(expected_drone_configs, actual_drone_config)
        self.assertEqual(expected_boundary_config, actual_boundary_config)
        self.assertEqual(expected_collision_config, acutal_collision_config)

    def test_parse_if_not_provided_configs_should_use_default_configs(self):
        config = "{}"
        with self.assertLogs(logging.getLogger()) as log:
            actual_drone_configs, actual_boundary_config, acutal_collision_config = ConfigParser.parse(config)
        expected_log = ["WARNING:root:'drones' missing when parsing configs, using default drone_config. " +
                        "Position estimation may be inaccurate.",
                        "WARNING:root:'boundary_config' missing when parsing configs, using unlimited boundary_config. " +
                        "Time and position will be unlimited.",
                        "WARNING:root:'collision_config' missing when parsing configs, using default collision_config."]
        self.assertEqual(expected_log, log.output)
        self.assertEqual({"DEFAULT": DefaultDroneConfig()}, actual_drone_configs)
        self.assertEqual(BoundaryConfig.no_limit(), actual_boundary_config)
        self.assertEqual(DefaultCollisionConfig(), acutal_collision_config)

    def test_parse_if_configs_missing_fields_should_use_default_value(self):
        config = """
            {
              "drones": [],
              "boundary_config": {
              },
              "collision_config": {
              }
            }
            """
        with self.assertLogs(logging.getLogger()) as log:
            actual_drone_configs, actual_boundary_config, actual_collision_config = ConfigParser.parse(config)
        expected_log = ["WARNING:root:'drones' is empty when parsing configs, using default drone_config. " +
                        "Position estimation may be inaccurate.",
                        "WARNING:root:'max_seconds' missing when parsing 'boundary_config', " +
                        "using default value inf. There will be no limit on 'max_seconds'.",
                        "WARNING:root:'max_x_meters' missing when parsing 'boundary_config', " +
                        "using default value inf. There will be no limit on 'max_x_meters'.",
                        "WARNING:root:'max_y_meters' missing when parsing 'boundary_config', " +
                        "using default value inf. There will be no limit on 'max_y_meters'.",
                        "WARNING:root:'max_z_meters' missing when parsing 'boundary_config', " +
                        "using default value inf. There will be no limit on 'max_z_meters'.",
                        "WARNING:root:'min_x_meters' missing when parsing 'boundary_config', " +
                        "using default value -inf. There will be no limit on 'min_x_meters'.",
                        "WARNING:root:'min_y_meters' missing when parsing 'boundary_config', " +
                        "using default value -inf. There will be no limit on 'min_y_meters'.",
                        "WARNING:root:'min_z_meters' missing when parsing 'boundary_config', " +
                        "using default value -inf. There will be no limit on 'min_z_meters'.",
                        "WARNING:root:'collision_meters' missing when parsing 'collision_config', " +
                        "using default value 0. There will be no limit on 'collision_meters'.",
                        "WARNING:root:'time_interval_seconds' missing when parsing 'collision_config', " +
                        "using default value 0.1.",
                        "WARNING:root:'confidence_threshold' missing when parsing 'collision_config', " +
                        "using default value 1.0."]
        self.assertEqual(expected_log, log.output)
        expected_drone_configs = {"DEFAULT": DefaultDroneConfig()}
        expected_boundary_config = BoundaryConfig(max_seconds=float("inf"), max_x_meters=float("inf"),
                                                  max_y_meters=float("inf"), max_z_meters=float("inf"),
                                                  min_x_meters=float("-inf"), min_y_meters=float("-inf"),
                                                  min_z_meters=float("-inf"))
        expected_collision_config = DefaultCollisionConfig()
        self.assertEqual(expected_drone_configs, actual_drone_configs)
        self.assertEqual(expected_boundary_config, actual_boundary_config)
        self.assertEqual(expected_collision_config, actual_collision_config)

    def test_parse_if_drone_object_missing_fields_should_use_default_value(self):
        config = """
            {
              "drones": [{}],
              "boundary_config": {
                "max_seconds": 100,
                "max_x_meters": 10,
                "max_y_meters": 20,
                "max_z_meters": 30,
                "min_x_meters": -10,
                "min_y_meters": -20,
                "min_z_meters": -30
              },
              "collision_config": {
                "collision_meters": 0.3,
                "time_interval_seconds": 0.5,
                "confidence_threshold": 0.9
              }
            }
            """
        with self.assertLogs(logging.getLogger()) as log:
            actual_drone_configs, actual_boundary_config, actual_collision_config = ConfigParser.parse(config)
        expected_log = ["WARNING:root:'name' missing when parsing object in 'drones', using default value 'DEFAULT'.",
                        "WARNING:root:'init_position' missing when parsing drone 'DEFAULT', " +
                        "using default value (0, 0, 0). Position estimation may be inaccurate.",
                        "WARNING:root:'speed_mps' missing when parsing drone 'DEFAULT', " +
                        "using default value 1. Position estimation may be inaccurate.",
                        "WARNING:root:'rotate_speed_dps' missing when parsing drone 'DEFAULT', " +
                        "using default value 90. Position estimation may be inaccurate.",
                        "WARNING:root:'takeoff_height_meters' missing when parsing drone 'DEFAULT', " +
                        "using default value 1. Position estimation may be inaccurate."]
        self.assertEqual(expected_log, log.output)
        expected_drone_configs = {"DEFAULT": DroneConfig(init_position=(0, 0, 0), speed_mps=1,
                                                         rotate_speed_dps=90, takeoff_height_meters=1)}
        expected_boundary_config = BoundaryConfig(max_seconds=100, max_x_meters=10, max_y_meters=20, max_z_meters=30,
                                                  min_x_meters=-10, min_y_meters=-20, min_z_meters=-30)
        expected_collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5,
                                                    confidence_threshold=0.9)
        self.assertEqual(expected_drone_configs, actual_drone_configs)
        self.assertEqual(expected_boundary_config, actual_boundary_config)
        self.assertEqual(expected_collision_config, actual_collision_config)

    def test_parse_if_init_position_missing_fields_should_use_default_value(self):
        config = """
            {
              "drones": [{
                "name": "DRONE1",
                "init_position": {},
                "speed_mps": 2,
                "rotate_speed_dps": 180,
                "takeoff_height_meters": 2
              }],
              "boundary_config": {
                "max_seconds": 100,
                "max_x_meters": 10,
                "max_y_meters": 20,
                "max_z_meters": 30,
                "min_x_meters": -10,
                "min_y_meters": -20,
                "min_z_meters": -30
              },
              "collision_config": {
                "collision_meters": 0.3,
                "time_interval_seconds": 0.5,
                "confidence_threshold": 0.9
              }
            }
            """

        with self.assertLogs(logging.getLogger()) as log:
            actual_drone_configs, actual_boundary_config, actual_collision_config = ConfigParser.parse(config)
        expected_log = ["WARNING:root:'x' missing when parsing drone 'DRONE1', using default value 0. " +
                        "Position estimation may be inaccurate.",
                        "WARNING:root:'y' missing when parsing drone 'DRONE1', using default value 0. " +
                        "Position estimation may be inaccurate.",
                        "WARNING:root:'z' missing when parsing drone 'DRONE1', using default value 0. " +
                        "Position estimation may be inaccurate."]
        self.assertEqual(expected_log, log.output)
        expected_drone_configs = {"DRONE1": DroneConfig(init_position=(0, 0, 0), speed_mps=2,
                                                        rotate_speed_dps=180, takeoff_height_meters=2)}
        expected_boundary_config = BoundaryConfig(max_seconds=100, max_x_meters=10, max_y_meters=20, max_z_meters=30,
                                                  min_x_meters=-10, min_y_meters=-20, min_z_meters=-30)
        expected_collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5,
                                                    confidence_threshold=0.9)
        self.assertEqual(expected_drone_configs, actual_drone_configs)
        self.assertEqual(expected_boundary_config, actual_boundary_config)
        self.assertEqual(expected_collision_config, actual_collision_config)

    def test_parse_if_name_is_empty_string_should_use_default_value(self):
        config = """
            {
              "drones": [{
                "name": "",
                "init_position": {"x": 1, "y": 2, "z": 3},
                "speed_mps": 2,
                "rotate_speed_dps": 180,
                "takeoff_height_meters": 2
              }],
              "boundary_config": {
                "max_seconds": 100,
                "max_x_meters": 10,
                "max_y_meters": 20,
                "max_z_meters": 30,
                "min_x_meters": -10,
                "min_y_meters": -20,
                "min_z_meters": -30
              },
              "collision_config": {
                "collision_meters": 0.3,
                "time_interval_seconds": 0.5,
                "confidence_threshold": 0.9
              }
            }
            """

        with self.assertLogs(logging.getLogger()) as log:
            actual_drone_configs, actual_boundary_config, actual_collision_config = ConfigParser.parse(config)
        expected_log = ["WARNING:root:'name' cannot be an empty string, using default value 'DEFAULT' instead."]
        self.assertEqual(expected_log, log.output)
        expected_drone_configs = {"DEFAULT": DroneConfig(init_position=(1, 2, 3), speed_mps=2,
                                                         rotate_speed_dps=180, takeoff_height_meters=2)}
        expected_boundary_config = BoundaryConfig(max_seconds=100, max_x_meters=10, max_y_meters=20, max_z_meters=30,
                                                  min_x_meters=-10, min_y_meters=-20, min_z_meters=-30)
        expected_collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5,
                                                    confidence_threshold=0.9)
        self.assertEqual(expected_drone_configs, actual_drone_configs)
        self.assertEqual(expected_boundary_config, actual_boundary_config)
        self.assertEqual(expected_collision_config, actual_collision_config)

    def test_parse_if_name_repeated_should_ignore_second_appearance(self):
        config = """
            {
              "drones": [{
                "name": "DRONE1",
                "init_position": {"x": 1, "y": 2, "z": 3},
                "speed_mps": 2,
                "rotate_speed_dps": 180,
                "takeoff_height_meters": 2
              },{
                "name": "DRONE1",
                "init_position": {"x": 4, "y": 5, "z": 6},
                "speed_mps": 1,
                "rotate_speed_dps": 90,
                "takeoff_height_meters": 1
              }],
              "boundary_config": {
                "max_seconds": 100,
                "max_x_meters": 10,
                "max_y_meters": 20,
                "max_z_meters": 30,
                "min_x_meters": -10,
                "min_y_meters": -20,
                "min_z_meters": -30
              },
              "collision_config": {
                "collision_meters": 0.3,
                "time_interval_seconds": 0.5,
                "confidence_threshold": 0.9
              }
            }
            """

        with self.assertLogs(logging.getLogger()) as log:
            actual_drone_configs, actual_boundary_config, actual_collision_config = ConfigParser.parse(config)
        expected_log = ["WARNING:root:Drone name 'DRONE1' appeared more than ones in 'drones', ignored."]
        self.assertEqual(expected_log, log.output)
        expected_drone_configs = {"DRONE1": DroneConfig(init_position=(1, 2, 3), speed_mps=2,
                                                        rotate_speed_dps=180, takeoff_height_meters=2)}
        expected_boundary_config = BoundaryConfig(max_seconds=100, max_x_meters=10, max_y_meters=20, max_z_meters=30,
                                                  min_x_meters=-10, min_y_meters=-20, min_z_meters=-30)
        expected_collision_config = CollisionConfig(collision_meters=0.3, time_interval_seconds=0.5,
                                                    confidence_threshold=0.9)
        self.assertEqual(expected_drone_configs, actual_drone_configs)
        self.assertEqual(expected_boundary_config, actual_boundary_config)
        self.assertEqual(expected_collision_config, actual_collision_config)



import socket
import netifaces
import netaddr
import threading
import time
from libs.TelloEduSwarmSearch.tello import Tello


class CommsManager:

    #
    # CLASS INIT & SETUP
    #

    def __init__(self):
        """ Open sockets ready for communicating with one or more Tellos.

            Also initiate the threads for receiving control messages and status from Tello.
            Also create the placeholder list for Tello objects.
        """

        self.terminate_comms = False

        # Socket for primary bi-directional communication with Tello
        self.control_port = 8889
        self.control_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # socket for sending cmd
        self.control_socket.bind(('', self.control_port))

        # Socket for receiving status messages from Tello - not activated here
        self.status_port = 8890
        self.status_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.status_thread = None

        # Thread for receiving messages from Tello
        self.receive_thread = threading.Thread(target=self._receive_thread)
        self.receive_thread.daemon = True
        self.receive_thread.start()

        # Reference to all active Tellos
        self.tellos = []

    def init_tellos(self, sn_list, get_status=False, first_ip=1, last_ip=254):
        """ Search the network until found the specified number of Tellos, then get each Tello ready for use.

            This must be run once; generally the first thing after initiating CommsManager.
            The 'command' message is sent to every IP on the network, with the response_handler thread managing the
             responses to create Tello objects in self.tellos.
            A command_handler is then created for each, which manages the command_queue for each.
            Finally, each Tello is queried for its serial number, which is stored in the Tello object with its number.

            :param sn_list: List of serial numbers, in order we want to number the Tellos.
            :param get_status: True to listen for and record the status messages from the Tellos.
            :param first_ip: If known, we can specify a smaller range of IP addresses to speed up the search.
            :param last_ip: If known, we can specify a smaller range of IP addresses to speed up the search.
        """

        # Get network addresses to search
        subnets, address = self._get_subnets()
        possible_addr = []

        # Create a list of possible IP addresses to search
        for subnet, netmask in subnets:
            for ip in netaddr.IPNetwork('%s/%s' % (subnet, netmask)):
                if not (first_ip <= int(str(ip).split('.')[3]) <= last_ip):
                    continue
                # Don't add the server's address to the list
                if str(ip) in address:
                    continue
                possible_addr.append(str(ip))

        # Continue looking until we've found them all
        num = len(sn_list)
        while len(self.tellos) < num:
            print('[Tello Search]Looking for %d Tello(s)' % (num - len(self.tellos)))

            # Remove any found Tellos from the list to search
            for tello_ip in [tello.ip for tello in self.tellos]:
                if tello_ip in possible_addr:
                    possible_addr.remove(tello_ip)

            # Try contacting Tello via each possible_addr
            for ip in possible_addr:
                self.control_socket.sendto('command'.encode(), (ip, self.control_port))

            # Responses to the command above will be picked up in receive_thread.  Here we check regularly to see if
            #  they've all been found, so we can break out quickly.  But after several failed attempts, go around the
            #  whole loop again and retry contacting.
            for _ in range(0, 10):
                time.sleep(0.5)
                if len(self.tellos) >= num:
                    break

        # Once we have all Tellos, startup a command_handler for each.  These manage the command queues for each Tello.
        for tello in self.tellos:
            command_handler_thread = threading.Thread(target=self._command_handler, args=(tello,))
            command_handler_thread.daemon = True
            command_handler_thread.start()

        # Start the status_handler, if needed.  This receives and constantly updates the status of each Tello.
        if get_status:
            self.status_socket.bind(('', self.status_port))
            self.status_thread = threading.Thread(target=self._status_thread)
            self.status_thread.daemon = True
            self.status_thread.start()

        # Query each Tello to get its serial number - saving the cmd_id so we can match-up responses when they arrive
        tello_cmd_id = []
        for tello in self.tellos:
            # Save the tello together with the returned cmd_id, so we can match the responses with the right Tello below
            tello_cmd_id.append((tello, tello.add_to_command_queue('sn?', 'Read', None)))

        # Assign the sn to each Tello, as responses become available.
        # Note that log_wait_response will block until the response is received.
        for tello, cmd_id in tello_cmd_id:
            tello.sn = tello.log_wait_response(cmd_id).response
            # Once we know the SN, look it up in the supplied sn_list and assign the correct tello_num
            for index, sn in enumerate(sn_list, 1):
                if tello.sn == sn:
                    tello.num = index

        # Sort the list of Tellos by their num
        self.tellos.sort(key=lambda tello: tello.num)

    #
    # PUBLIC METHODS
    #

    def queue_command(self, command, command_type, tello_num, on_error=None):
        """ Add a new command to the Tello's (either one Tello or all) command queue - returning the cmd_id.

            Note that if a Tello is marked as flight_completed, it will return -1 as its cmd_id.  These are not
             added to the list returned here, so can effectively be ignored by calling functions.

            :param command: The Tello SDK string (e.g. 'forward 50' or 'battery?') to send to the Tello(s).
            :param command_type: Either 'Control', 'Set' or 'Read' - corresponding to the Tello SDK documentation.
            :param tello_num: Either 'All' or a Tello number (1,2,...)
            :param on_error: A different Tello SDK string to be sent if command returns an error.
            :return: A list of tuples in the form [(tello_num, cmd_id),...].
        """
        # Determine which Tellos to use, and add the command to the appropriate Tello's queue.
        cmd_ids = []
        if tello_num == 'All':
            for tello in self.tellos:
                # If command is for all tellos, send to each and save the cmd_id in a list
                cmd_id = tello.add_to_command_queue(command, command_type, on_error)
                if cmd_id != -1:
                    cmd_ids.append((tello.num, cmd_id))
        else:
            tello = self.get_tello(num=tello_num)
            cmd_id = tello.add_to_command_queue(command, command_type, on_error)
            if cmd_id != -1:
                cmd_ids.append((tello.num, cmd_id))
        return cmd_ids

    def wait_sync(self):
        """ Used to pause the main thread whilst all Tellos catch up, to bring all Tellos into sync.

            Simply checks with each Tello object that each individually has fully processed its queue and responses.
            The wait_until_idle command is a blocking function, so won't return until ready.
        """
        for tello in self.tellos:
            tello.wait_until_idle()

    def get_tello(self, num):
        """ Shortcut function to return a specific Tello instance, based on its number.

            :param num: Tello number, as an integer (e.g. 1,2,...)
            :return: Tello object
        """
        for tello in self.tellos:
            if tello.num == num:
                return tello
        raise RuntimeError('Tello not found!')

    def close_connections(self):
        """ Close all comms - to tidy up before exiting """
        self.terminate_comms = True
        self.control_socket.close()
        self.status_socket.close()

    #
    # PRIVATE HELPER METHODS
    #

    @staticmethod
    def _get_subnets():
        """ Get the local subnet and server IP address """
        subnets = []
        addr_list = []
        ifaces = netifaces.interfaces()
        for this_iface in ifaces:
            addrs = netifaces.ifaddresses(this_iface)

            if socket.AF_INET not in addrs:
                continue

            # Get IPv4 info
            ip_info = addrs[socket.AF_INET][0]
            address = ip_info['addr']
            netmask = ip_info['netmask']

            # Avoid searching when on very large subnets
            if netmask != '255.255.255.0':
                continue

            # Create IP object and get the network details
            # Note CIDR is a networking term, describing the IP/subnet address format
            cidr = netaddr.IPNetwork('%s/%s' % (address, netmask))
            network = cidr.network
            subnets.append((network, netmask))
            addr_list.append(address)
        return subnets, addr_list

    def _get_tello(self, ip):
        """ Private function to return the Tello object with the matching IP address.

            :param ip: IP address of the requested Tello object, as a string e.g. '123.45.678.90'
            :return: Tello object
        """
        for tello in self.tellos:
            if tello.ip == ip:
                return tello
        raise RuntimeError('Tello not found!')

    def _send_command(self, tello, cmd_id, command, command_type, on_error, timeout=10):
        """ Actually send a command to the Tello at specified IP address, recording details in the Tello's log.

            :param tello: The Tello object for which we're sending the command
            :param cmd_id: Corresponds to the id first given when in the Tello's queue, to be transferred to its log.
            :param command: The actual command from Tello SDK, e.g. 'battery?', 'forward 50', etc...
            :param command_type: Either 'Control', 'Set' or 'Read' - corresponding to the Tello SDK documentation.
            :param on_error: A different Tello SDK string to be sent if command returns an error.
        """

        # Add the command to the Tello's log first
        log_entry = tello.add_to_log(cmd_id, command, command_type, on_error)

        # Then send the command
        self.control_socket.sendto(command.encode(), (tello.ip, self.control_port))
        print('[Command  %s]Sent cmd: %s' % (tello.ip, command))

        # Wait until a response has been received, and handle timeout
        time_sent = time.time()
        while log_entry.response is None:
            now = time.time()
            if now - time_sent > timeout:
                print('[Command  %s]Failed to send: %s' % (tello.ip, command))
                log_entry.success = False
                log_entry.response = ''
                if log_entry.on_error is not None:
                    tello.add_to_command_queue(log_entry.on_error, log_entry.command_type, None)
                    print('[Command  %s]Queuing alternative cmd: %s' % (tello.ip, log_entry.on_error))
                return
            # Sleep briefly at the end of each loop, to prevent excessive CPU usage
            time.sleep(0.01)

    #
    # THREADS
    #

    def _command_handler(self, tello):
        """ Run Command Handler as a separate thread for each Tello, to manage the queue of commands.

            This runs as a separate thread so that applications can instantly add commands to multiple queues
            simultaneously, and then each of these threads (one per Tello) can all actually send the command
            together.  The send_command function called from here is a blocking function, which doesn't return
            until the response has been received or the command exceeds its timeout.

            :param tello: The Tello object with which the command_handler should be associated.
        """
        while True:
            # If nothing in the queue, just keep looping
            while not tello.command_queue:
                time.sleep(0.01)
            # Pop command off the Tello's queue, then send the command.
            # Note as part of send_command the same details will be added back into Tello's log.
            command = tello.command_queue.pop(0)
            self._send_command(tello, command.cmd_id, command.command, command.command_type, command.on_error)

    def _receive_thread(self):
        """ Listen continually to responses from the Tello - should run in its own thread.

            This method includes capturing and saving each Tello the first time it responds.
            If it is a known Tello, the response will be matched against the Tello's log, always recording the response
            against the last log entry as commands sent to each Tello are strictly sequential.
            Responses are also tested for success or failure, and if relevant an alternative command may be sent
            immediately on error.
        """

        while not self.terminate_comms:
            try:
                # Get responses from all Tellos - this line blocks until a message is received - and reformat values
                response, ip = self.control_socket.recvfrom(1024)
                response = response.decode().strip()
                ip = str(ip[0])

                # Capture Tellos when they respond for the first time
                if response.lower() == 'ok' and ip not in [tello.ip for tello in self.tellos]:
                    print('[Tello Search]Found Tello on IP %s' % ip)
                    self.tellos.append(Tello(ip))
                    continue

                # Get the current log entry for this Tello
                tello = self._get_tello(ip)
                log_entry = tello.log_entry()

                # Determine if the response was ok / error (or reading a value)
                send_on_error = False
                if log_entry.command_type in ['Control', 'Set']:
                    if response == 'ok':
                        log_entry.success = True
                    else:
                        log_entry.success = False
                        if log_entry.on_error is not None:
                            # If this command wasn't successful, and there's an on_error entry, flag to send it later.
                            send_on_error = True
                elif log_entry.command_type == 'Read':
                    # Assume Read commands are always successful... not aware they can return anything else!?
                    log_entry.success = True
                else:
                    print('[Response %s]Invalid command_type: %s' % (ip, log_entry.command_type))
                # Save .response *after* .success, as elsewhere we use .response as a check to move on - avoids race
                # conditions across the other running threads, which might otherwise try to use .success before saved.
                log_entry.response = response
                print('[Response %s]Received: %s' % (ip, response))
                # If required, queue the alternative command - assume same command type as the original.
                if send_on_error:
                    tello.add_to_command_queue(log_entry.on_error, log_entry.command_type, None)
                    print('[Command  %s]Queuing alternative cmd: %s' % (ip, log_entry.on_error))

            except socket.error as exc:
                if not self.terminate_comms:
                    # Report socket errors, but only if we've not told it to terminate_comms.
                    print('[Socket Error]Exception socket.error : %s' % exc)

    def _status_thread(self):
        """ Listen continually to status from the Tellos - should run in its own thread.

            Listens for status messages from each Tello, and saves them in the Tello object as they arrive.
        """

        while not self.terminate_comms:
            try:
                response, ip = self.status_socket.recvfrom(1024)
                response = response.decode()
                if response == 'ok':
                    continue
                ip = ''.join(str(ip[0]))
                tello = self._get_tello(ip)
                tello.status.clear()
                status_parts = response.split(';')
                for status_part in status_parts:
                    key_value = status_part.split(':')
                    if len(key_value) == 2:
                        tello.status[key_value[0]] = key_value[1]

            except socket.error as exc:
                if not self.terminate_comms:
                    # Report socket errors, but only if we've not told it to terminate_comms.
                    print('[Socket Error]Exception socket.error : %s' % exc)



import unittest

from xdrone.compiler.compiler_utils.drones import NullDrone
from xdrone.compiler.compiler_utils.type import Type, ListType, EmptyList


class TestTypes(unittest.TestCase):
    def test_int(self):
        self.assertEqual(Type.int(), Type.int())
        self.assertEqual("int", Type.int().type_name)
        self.assertEqual(0, Type.int().default_value)
        self.assertEqual("int", str(Type.int()))

    def test_decimal(self):
        self.assertEqual(Type.decimal(), Type.decimal())
        self.assertEqual("decimal", Type.decimal().type_name)
        self.assertEqual(0.0, Type.decimal().default_value)
        self.assertEqual("decimal", str(Type.decimal()))

    def test_string(self):
        self.assertEqual(Type.string(), Type.string())
        self.assertEqual("string", Type.string().type_name)
        self.assertEqual("", Type.string().default_value)
        self.assertEqual("string", str(Type.string()))

    def test_boolean(self):
        self.assertEqual(Type.boolean(), Type.boolean())
        self.assertEqual("boolean", Type.boolean().type_name)
        self.assertEqual(False, Type.boolean().default_value)
        self.assertEqual("boolean", str(Type.boolean()))

    def test_vector(self):
        self.assertEqual(Type.vector(), Type.vector())
        self.assertEqual("vector", Type.vector().type_name)
        self.assertEqual([0.0, 0.0, 0.0], Type.vector().default_value)
        self.assertEqual("vector", str(Type.vector()))

    def test_drone(self):
        self.assertEqual(Type.drone(), Type.drone())
        self.assertEqual("drone", Type.drone().type_name)
        self.assertEqual(NullDrone(), Type.drone().default_value)
        self.assertEqual("drone", str(Type.drone()))

    def test_list(self):
        list_type = Type.list_of(Type.int())
        self.assertEqual(Type.list_of(Type.int()), list_type)
        self.assertEqual(ListType(Type.int()), list_type)
        self.assertEqual("list[int]", list_type.type_name)
        self.assertEqual([], list_type.default_value)
        self.assertEqual(Type.int(), list_type.elem_type)
        self.assertEqual("list[int]", str(list_type))

    def test_nested_list(self):
        inner = Type.list_of(Type.int())
        outer = Type.list_of(inner)
        self.assertEqual(ListType(Type.list_of(Type.int())), outer)
        self.assertEqual("list[list[int]]", outer.type_name)
        self.assertEqual([], outer.default_value)
        self.assertEqual(Type.list_of(Type.int()), outer.elem_type)
        self.assertEqual("list[list[int]]", str(outer))

    def test_empty_list(self):
        empty_list = Type.empty_list()
        self.assertEqual(Type.empty_list(), empty_list)
        self.assertEqual(EmptyList(), empty_list)
        self.assertEqual("list[]", empty_list.type_name)
        self.assertEqual([], empty_list.default_value)
        self.assertEqual(Type("all", 0), empty_list.elem_type)
        self.assertEqual("list[]", str(empty_list))

    def test_eq(self):
        types1 = [None, Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                  Type.list_of(Type.int()), Type.list_of(Type.decimal()),
                  Type.list_of(Type.list_of(Type.int())), Type.list_of(Type.list_of(Type.decimal()))]
        types2 = [None, Type.int(), Type.decimal(), Type.string(), Type.boolean(), Type.vector(), Type.drone(),
                  Type.list_of(Type.int()), Type.list_of(Type.decimal()),
                  Type.list_of(Type.list_of(Type.int())), Type.list_of(Type.list_of(Type.decimal()))]
        for i in range(len(types1)):
            for j in range(len(types2)):
                if i == j:
                    self.assertEqual(types1[i], types2[j])
                else:
                    self.assertNotEqual(types1[i], types2[j])

        self.assertEqual(Type.empty_list(), Type.empty_list())
        for type in types1:
            if isinstance(type, ListType):
                self.assertEqual(Type.empty_list(), type)
                self.assertEqual(type, Type.empty_list())
            else:
                self.assertNotEqual(Type.empty_list(), type)
                self.assertNotEqual(type, Type.empty_list())

    def test_corrupted_type_should_not_equal_to_list_type(self):
        self.assertNotEqual(Type("list[int]", []), Type.list_of(Type.int()))
        self.assertNotEqual(Type("list[int]", []), ListType(Type.int()))

    def test_corrupted_type_should_not_affect_correct_type(self):
        int_type = Type.int()
        corrupted_type = Type.int()
        corrupted_type._type_name = "corrupted"
        self.assertNotEqual(int_type, corrupted_type)
        self.assertEqual("int", str(Type.int()))
        self.assertEqual(0, Type.int().default_value)
        self.assertEqual("corrupted", str(corrupted_type))
        self.assertEqual(0, corrupted_type.default_value)



from __future__ import annotations

import copy
from typing import Union, List, Set


class Command:
    def __init__(self, opcode: str, operands: list):
        self._opcode = opcode
        self._operands = operands

    @property
    def opcode(self) -> str:
        return copy.deepcopy(self._opcode)

    @property
    def operands(self) -> list:
        return copy.deepcopy(self._operands)

    @staticmethod
    def takeoff() -> Command:
        return Command("takeoff", [])

    @staticmethod
    def land() -> Command:
        return Command("land", [])

    @staticmethod
    def up(distance_meters: Union[int, float]) -> Command:
        return Command("up", [distance_meters])

    @staticmethod
    def down(distance_meters: Union[int, float]) -> Command:
        return Command("down", [distance_meters])

    @staticmethod
    def left(distance_meters: Union[int, float]) -> Command:
        return Command("left", [distance_meters])

    @staticmethod
    def right(distance_meters: Union[int, float]) -> Command:
        return Command("right", [distance_meters])

    @staticmethod
    def forward(distance_meters: Union[int, float]) -> Command:
        return Command("forward", [distance_meters])

    @staticmethod
    def backward(distance_meters: Union[int, float]) -> Command:
        return Command("backward", [distance_meters])

    @staticmethod
    def rotate_left(angle_degrees: Union[int, float]) -> Command:
        return Command("rotate_left", [angle_degrees])

    @staticmethod
    def rotate_right(angle_degrees: Union[int, float]) -> Command:
        return Command("rotate_right", [angle_degrees])

    @staticmethod
    def wait(time_seconds: Union[int, float]) -> Command:
        return Command("wait", [time_seconds])

    def __str__(self):
        return "Command: {{ opcode: {}, operands: {} }}".format(self._opcode, self._operands)

    def __eq__(self, other):
        if isinstance(other, Command):
            return other._opcode == self._opcode and other._operands == self._operands
        return False


class AbstractDroneCommand:
    def get_drones_involved(self) -> Set[str]:
        raise NotImplementedError("get_drones_involved not supported in AbstractDroneCommand")

    def to_command_str(self) -> str:
        raise NotImplementedError("to_command_str not supported in AbstractDroneCommand")


class SingleDroneCommand(AbstractDroneCommand):
    def __init__(self, drone_name: str, command: Command):
        self._drone_name = drone_name
        self._command = command

    @property
    def drone_name(self) -> str:
        return copy.deepcopy(self._drone_name)

    @property
    def command(self) -> Command:
        return copy.deepcopy(self._command)

    def get_drones_involved(self) -> Set[str]:
        return {self.drone_name}

    def to_command_str(self) -> str:
        return "{}.{}{};".format(self._drone_name,
                                 self._command.opcode,
                                 str(self._command.operands).replace("[", "(").replace("]", ")"))

    def __str__(self):
        return "SingleDroneCommand: {{ drone_name: {}, command: {} }}".format(self._drone_name, self._command)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, SingleDroneCommand):
            return other._drone_name == self._drone_name and other._command == self._command
        return False


class RepeatDroneNameException(Exception):
    def __init__(self, repeated_names: Set[str]):
        self._repeated_names = repeated_names

    @property
    def repeated_names(self) -> Set[str]:
        return copy.deepcopy(self._repeated_names)


class ParallelDroneCommands(AbstractDroneCommand):
    def __init__(self, branches: List[List[AbstractDroneCommand]] = None):
        self._branches = []
        self._drones_involved_each_branch = []
        if branches is not None:
            for branch in branches:
                self.add(branch)

    @property
    def branches(self) -> List[List[AbstractDroneCommand]]:
        return copy.deepcopy(self._branches)

    @property
    def drones_involved_each_branch(self) -> List[Set[str]]:
        return copy.deepcopy(self._drones_involved_each_branch)

    def get_drones_involved(self) -> Set[str]:
        return set.union(set(), *self._drones_involved_each_branch)

    def add(self, drone_commands: List[AbstractDroneCommand]):
        drones_to_be_involved = set.union(set(),
                                          *[drone_command.get_drones_involved() for drone_command in drone_commands])
        repeated_names = self.get_drones_involved().intersection(drones_to_be_involved)
        if len(repeated_names) > 0:
            raise RepeatDroneNameException(repeated_names)
        self._branches.append(drone_commands)
        self._drones_involved_each_branch.append(drones_to_be_involved)

    def to_command_str(self) -> str:
        result = "{ "
        for i, commands in enumerate(self._branches):
            for command in commands:
                result += command.to_command_str() + " "
            result += "} || { " if i < len(self._branches) - 1 else "};"
        return result

    def __str__(self):
        result = "ParallelDroneCommands: { "
        result += ", ".join([str(commands) for commands in self._branches])
        result += " }"
        return result

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, ParallelDroneCommands):
            return other._branches == self._branches
        return False



import logging
import multiprocessing
import urllib.parse
import webbrowser

import click

from xdrone.command_generators import generate_commands_with_config
from xdrone.command_converters.dji_tello_edu_drone_executor import DJITelloEduExecutor
from xdrone.command_converters.simulation_converter import SimulationConverter


@click.command()
@click.option('--validate', 'function', flag_value='validate', help='Validate your code.', default=True)
@click.option('--simulate', 'function', flag_value='simulate', help='Simulate the drone flight.')
@click.option('--fly', 'function', flag_value='fly', help='Fly the drone.')
@click.option('--code', type=click.Path(), help='The file contains your code.')
@click.option('--config', type=click.Path(), help='The configuration json of your drone and flight environment.')
@click.option('--timeout', default=10, type=click.INT, help='Timeout of compilation.')
@click.option('--port', default=8080, type=click.INT, help='Port on localhost where the simulator server is running.')
@click.option('--no-check', is_flag=True, help='No safety check will be performed.')
@click.option('--save-report', is_flag=True, help='Save the report of the collision check.')
def xdrone(function, code, config, timeout, port, no_check, save_report):
    if code is None:
        code = click.prompt("Please enter path to your code", type=click.Path())
    if config is None:
        config = click.prompt("Please enter path to your configuration", type=click.Path())

    with open(code, mode='r') as file:
        program = file.read()
    with open(config, mode='r') as file:
        config = file.read()

    if no_check:
        if not click.confirm("Safety checks will be skipped, are you sure?"):
            print("Aborted.")
            return

    has_checks = not no_check
    queue = multiprocessing.Queue()
    _run_with_timeout(_validate, (program, config, has_checks, save_report, queue), timeout,
                      "Timed out. Please check whether your code contains an infinite loop. " +
                      "Your can change the timeout time by adding the '--timeout' tag.")
    if queue.empty():
        return
    drone_commands, drone_config_map = queue.get()
    if function == "simulate":
        _simulate(drone_commands, drone_config_map, port)
    if function == "fly":
        _fly(drone_commands, drone_config_map)


def _run_with_timeout(target, args, timeout, timeout_msg="Timeout"):
    p = multiprocessing.Process(target=target, args=args)
    p.start()
    p.join(timeout)
    if p.is_alive():
        print(timeout_msg)
        p.terminate()
        p.join()


def _validate(program, config, has_checks, save_report, queue):
    print("Validating your program...")
    if not has_checks:
        print("Skipping safety checks... only syntax will be checked")
    if save_report:
        print("Reports will be saved at directory ./reports/")
    try:
        drone_commands, drone_config_map, _ = generate_commands_with_config(program, config, has_checks, save_report)
        print("Your program is valid.")
        queue.put((drone_commands, drone_config_map))
    except Exception as e:
        print("Failed to validate your program, error: " + str(e))


def _simulate(drone_commands, drone_config_map, port):
    print("Start simulation...")
    if not _is_port_in_use(port):
        print("Aborted. Please make sure xDrone Simulator is running on localhost port {}. ".format(port) +
              "Your can change the port by adding the '--port' tag.")
        return
    simulation_json = SimulationConverter().convert(drone_commands, drone_config_map)
    url = "http://localhost:{}/?data={}".format(port, urllib.parse.quote(simulation_json))
    print(url)
    webbrowser.open(url)


def _is_port_in_use(port):
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def _fly(drone_commands, drone_config_map):
    # {"DRONE1": "0TQDG19EDB26V6", "DRONE2": "0TQDG19EDBNC96"}
    name_id_map = {}
    for name in drone_config_map.keys():
        id = click.prompt("Enter your Tello EDU drone id for \"{}\"".format(name), type=click.STRING)
        name_id_map[name] = id

    print("Start to fly your drone...")
    DJITelloEduExecutor(name_id_map).execute_drone_commands(drone_commands)


def main():
    logging.basicConfig(level=logging.INFO)
    try:
        xdrone(standalone_mode=False)
    except Exception as e:
        print("Aborted. Error: {}".format(str(e)))
        return


if __name__ == '__main__':
    main()



from __future__ import annotations

from xdrone.shared.safety_check_error import SafetyCheckError
from xdrone.shared.state import State


class BoundaryConfig:
    def __init__(self, max_seconds: float = float('inf'),
                 max_x_meters: float = float('inf'), max_y_meters: float = float('inf'),
                 max_z_meters: float = float('inf'), min_x_meters: float = float('-inf'),
                 min_y_meters: float = float('-inf'), min_z_meters: float = float('-inf')):
        if max_seconds < 0:
            raise ValueError("max_seconds should >= 0")
        if max_x_meters < min_x_meters:
            raise ValueError("max_x_meters should >= min_x_meters")
        if max_y_meters < min_y_meters:
            raise ValueError("max_y_meters should >= min_y_meters")
        if max_z_meters < min_z_meters:
            raise ValueError("max_z_meters should >= min_z_meters")
        self._max_seconds = max_seconds
        self._max_x_meters = max_x_meters
        self._max_y_meters = max_y_meters
        self._max_z_meters = max_z_meters
        self._min_x_meters = min_x_meters
        self._min_y_meters = min_y_meters
        self._min_z_meters = min_z_meters

    @staticmethod
    def no_limit() -> BoundaryConfig:
        return BoundaryConfig()

    def __str__(self):
        return ("BoundaryConfig: {{ max_seconds: {}, x_range_meters: {}, y_range_meters: {}, z_range_meters: {} }}"
                .format(self._max_seconds, (self._min_x_meters, self._max_x_meters),
                        (self._min_y_meters, self._max_y_meters), (self._min_z_meters, self._max_z_meters)))

    def __eq__(self, other):
        if isinstance(other, BoundaryConfig):
            return other._max_seconds == self._max_seconds and other._max_x_meters == self._max_x_meters \
                   and other._max_y_meters == self._max_y_meters and other._max_z_meters == self._max_z_meters \
                   and other._min_x_meters == self._min_x_meters and other._min_y_meters == self._min_y_meters \
                   and other._min_z_meters == self._min_z_meters
        return False

    def check_state(self, drone_name: str, state: State) -> None:
        if state.x_meters > self._max_x_meters:
            raise SafetyCheckError("Drone '{}': the x coordinate {} will go beyond its upper limit {}"
                                   .format(drone_name, state.x_meters, self._max_x_meters))
        if state.y_meters > self._max_y_meters:
            raise SafetyCheckError("Drone '{}': the y coordinate {} will go beyond its upper limit {}"
                                   .format(drone_name, state.y_meters, self._max_y_meters))
        if state.z_meters > self._max_z_meters:
            raise SafetyCheckError("Drone '{}': the z coordinate {} will go beyond its upper limit {}"
                                   .format(drone_name, state.z_meters, self._max_z_meters))
        if state.x_meters < self._min_x_meters:
            raise SafetyCheckError("Drone '{}': the x coordinate {} will go beyond its lower limit {}"
                                   .format(drone_name, state.x_meters, self._min_x_meters))
        if state.y_meters < self._min_y_meters:
            raise SafetyCheckError("Drone '{}': the y coordinate {} will go beyond its lower limit {}"
                                   .format(drone_name, state.y_meters, self._min_y_meters))
        if state.z_meters < self._min_z_meters:
            raise SafetyCheckError("Drone '{}': the z coordinate {} will go beyond its lower limit {}"
                                   .format(drone_name, state.z_meters, self._min_z_meters))
        if state.time_used_seconds > self._max_seconds:
            raise SafetyCheckError("Drone '{}': the time used {} seconds will go beyond the time limit {} seconds"
                                   .format(drone_name, state.time_used_seconds, self._max_seconds))
